@@ -4,37 +4,19 @@ import asyncio
 import logging
 import socket
 from asyncio import Lock
-from typing import Callable, Coroutine, Any
+from typing import Callable, Coroutine
 
 import async_timeout
 
 from roborock.api import RoborockClient, SPECIAL_COMMANDS
 from roborock.containers import RoborockLocalDeviceInfo
-from roborock.exceptions import RoborockTimeout, CommandVacuumError, RoborockConnectionException
-from roborock.typing import RoborockCommand
+from roborock.exceptions import RoborockTimeout, CommandVacuumError, RoborockConnectionException, RoborockException
+from roborock.roborock_message import RoborockParser, RoborockMessage
+from roborock.typing import RoborockCommand, CommandInfoMap
 from roborock.util import get_running_loop_or_create_one
 
-secured_prefix = 199
-get_prefix = 119
-app_prefix = 135
-set_prefix = 151
-nop_prefix = 21
-
 _LOGGER = logging.getLogger(__name__)
 
-APP_COMMANDS = (
-    RoborockCommand.GET_MULTI_MAPS_LIST,
-    RoborockCommand.GET_CLEAN_RECORD,
-    RoborockCommand.GET_DUST_COLLECTION_MODE,
-    RoborockCommand.GET_WASH_TOWEL_MODE,
-    RoborockCommand.GET_SMART_WASH_PARAMS,
-    "app_"
-)
-
-SET_COMMANDS = (
-    "set_"
-)
-
 
 class RoborockLocalClient(RoborockClient):
 
@@ -42,10 +24,16 @@ class RoborockLocalClient(RoborockClient):
         super().__init__("abc", devices_info)
         self.loop = get_running_loop_or_create_one()
         self.device_listener: dict[str, RoborockSocketListener] = {
-            device_id: RoborockSocketListener(device_info.network_info.ip, device_id, self.on_message)
+            device_id: RoborockSocketListener(
+                device_info.network_info.ip,
+                device_info.device.local_key,
+                self.on_message
+            )
             for device_id, device_info in devices_info.items()
         }
         self._mutex = Lock()
+        self._batch_structs: list[RoborockMessage] = []
+        self._executing = False
 
     async def async_connect(self):
         await asyncio.gather(*[
@@ -53,47 +41,62 @@ class RoborockLocalClient(RoborockClient):
             for listener in self.device_listener.values()
         ])
 
-    async def async_disconnect(self) -> Any:
+    async def async_disconnect(self) -> None:
         await asyncio.gather(*[listener.disconnect() for listener in self.device_listener.values()])
 
-    # async def test(
-    #         self, device_id: str
-    # ):
-    #     request_id, timestamp, payload = self._get_payload(RoborockCommand.NOP)
-    #     prefix = 21
-    #     request_protocol = 0
-    #     msg = self._encode_msg(device_id, request_id, request_protocol, timestamp, payload, prefix)
-    #     # Send the command to the Roborock device
-    #     listener = self.device_listener.get(device_id)
-    #     await listener.send_message(msg)
-    #     response_protocol = 1
-    #     (response, err) = await self._async_response(request_id, response_protocol)
-    #     if err:
-    #         raise CommandVacuumError('NOP', err) from err
-    #     _LOGGER.debug(f"id={request_id} Response from NOP: {response}")
-    #     return response
+    def build_roborock_message(
+            self, method: RoborockCommand, params: list = None
+    ) -> RoborockMessage:
+        secured = True if method in SPECIAL_COMMANDS else False
+        request_id, timestamp, payload = self._get_payload(method, params, secured)
+        _LOGGER.debug(f"id={request_id} Requesting method {method} with {params}")
+        command_info = CommandInfoMap.get(method)
+        if not command_info:
+            raise RoborockException(f"Request {method} have unknown prefix. Can't execute in offline mode")
+        prefix = CommandInfoMap.get(method).prefix
+        request_protocol = 4
+        return RoborockMessage(
+            prefix=prefix,
+            timestamp=timestamp,
+            protocol=request_protocol,
+            payload=payload
+        )
 
     async def send_command(
             self, device_id: str, method: RoborockCommand, params: list = None
     ):
-        secured = True if method in SPECIAL_COMMANDS else False
-        request_id, timestamp, payload = self._get_payload(method, params, secured)
-        _LOGGER.debug(f"id={request_id} Requesting method {method} with {params}")
-        prefix = secured_prefix if method in SPECIAL_COMMANDS \
-            else app_prefix if method.startswith(APP_COMMANDS) \
-            else set_prefix if method.startswith(SET_COMMANDS) \
-            else get_prefix
-        request_protocol = 4
-        msg = self._encode_msg(device_id, request_id, request_protocol, timestamp, payload, prefix)
+        roborock_message = self.build_roborock_message(method, params)
+        response = (await self.send_message(device_id, roborock_message))[0]
+        if isinstance(response, BaseException):
+            raise response
+        return response
+
+    async def async_local_response(self, roborock_message: RoborockMessage):
+        request_id = roborock_message.get_request_id()
+        if request_id is not None:
+            # response_protocol = 5 if roborock_message.prefix == secured_prefix else 4
+            response_protocol = 4
+            (response, err) = await self._async_response(request_id, response_protocol)
+            if err:
+                raise CommandVacuumError("", err) from err
+            _LOGGER.debug(f"id={request_id} Response from {roborock_message.get_method()}: {response}")
+            return response
+
+    async def send_message(
+            self, device_id: str, roborock_messages: list[RoborockMessage] | RoborockMessage
+    ):
+        if isinstance(roborock_messages, RoborockMessage):
+            roborock_messages = [roborock_messages]
+        local_key = self.devices_info[device_id].device.local_key
+        msg = RoborockParser.encode(roborock_messages, local_key)
         # Send the command to the Roborock device
         listener = self.device_listener.get(device_id)
+        _LOGGER.debug(f"Requesting device with {roborock_messages}")
         await listener.send_message(msg)
-        response_protocol = 5 if prefix == secured_prefix else 4
-        (response, err) = await self._async_response(request_id, response_protocol)
-        if err:
-            raise CommandVacuumError(method, err) from err
-        _LOGGER.debug(f"id={request_id} Response from {method}: {response}")
-        return response
+
+        return await asyncio.gather(
+            *[self.async_local_response(roborock_message) for roborock_message in roborock_messages],
+            return_exceptions=True)
 
 
 class RoborockSocket(socket.socket):
@@ -107,10 +110,10 @@ class RoborockSocket(socket.socket):
 class RoborockSocketListener:
     roborock_port = 58867
 
-    def __init__(self, ip: str, device_id: str, on_message: Callable[[str, bytes], Coroutine[None] | None],
+    def __init__(self, ip: str, local_key: str, on_message: Callable[[list[RoborockMessage]], Coroutine[None] | None],
                  timeout: float | int = 4):
         self.ip = ip
-        self.device_id = device_id
+        self.local_key = local_key
         self.socket = RoborockSocket(socket.AF_INET, socket.SOCK_STREAM)
         self.socket.setblocking(False)
         self.loop = get_running_loop_or_create_one()
@@ -118,13 +121,19 @@ class RoborockSocketListener:
         self.timeout = timeout
         self.is_connected = False
         self._mutex = Lock()
+        self.remaining = b''
 
     async def _main_coro(self):
         while not self.socket.is_closed:
             try:
                 message = await self.loop.sock_recv(self.socket, 4096)
                 try:
-                    await self.on_message(self.device_id, message)
+                    if self.remaining:
+                        message = self.remaining + message
+                        self.remaining = b''
+                    (parser_msg, remaining) = RoborockParser.decode(message, self.local_key)
+                    self.remaining = remaining
+                    await self.on_message(parser_msg)
                 except Exception as e:
                     _LOGGER.exception(e)
             except BrokenPipeError as e:
@@ -158,6 +167,7 @@ class RoborockSocketListener:
                 async with async_timeout.timeout(self.timeout):
                     await self.loop.sock_sendall(self.socket, data)
         except (asyncio.TimeoutError, asyncio.CancelledError):
+            await self.disconnect()
             raise RoborockTimeout(
                 f"Timeout after {self.timeout} seconds waiting for response"
             ) from None
